/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * END COPYRIGHT BLOCK **/
/*
 * buffer.c: Code for dealing with I/O buffers
 * 
 * See buffer.h for public documentation.
 *
 * Rob McCool
 * 
 */

#include "private/pprio.h"     /* for nspr20 - yjh */
#include "base/buffer.h"
#include "base/dbtbase.h"


/* ----------------------------- buffer_open ------------------------------ */

#ifdef FILE_UNIX_MMAP
NSPR_BEGIN_EXTERN_C
#include <sys/mman.h>           /* mmap, munmap */
NSPR_END_EXTERN_C
#endif

/* FUNCTION: filebuf_create
 * DESCRIPTION:
 *    A convenience function to create a file buffer if the file is already
 *    open or mmap'd.  It creates a new filebuf for use with the mmap'd file.
 *    If mmap_ptr is NULL, it creates a buffer with size bufsz.
 * INPUTS:
 *    fd - the file to buffer
 *    mmap_ptr - a pointer to the mmap'd region for this file.
 *    mmap_len - the length of the mmap'd region
 *    bufsz    - buffer size to use.
 * OUTPUTS:
 *    none
 * RETURNS:
 *    NULL on failure
 *    a valid pointer on success
 * RESTRICTIONS:
 */
filebuf_t *filebuf_create(SYS_FILE fd, caddr_t mmap_ptr, int mmap_len, int bufsz)
{
    filebuf_t *buf = (filebuf_t *)MALLOC(sizeof(filebuf_t));

    if (buf) {
        buf->fd = fd;
#ifdef FILE_MMAP
        buf->pos = 0;
#ifdef FILE_UNIX_MMAP
        buf->fp = mmap_ptr;
#else
        buf->fp = (char *)mmap_ptr;
#endif
        buf->len = mmap_len;
        buf->cursize = mmap_len;
#else /* FILE_MMAP */
        buf->pos = bufsz;
        buf->maxsize = bufsz;
        buf->cursize = bufsz;
#endif /* FILE_MMAP */
        buf->inbuf = NULL;
        buf->errmsg = NULL;
    }

    return buf;
}

/* FUNCTION: filebuf_close_buffer
 * DESCRIPTION:
 *    Deallocates a buffer but does not close the underlying file.  If 
 *    clean_mmap is set to 1, and the file is mmap'd, the mmap'd region
 *    will also be closed.  But the file itself will remain open.
 * INPUTS:
 *    buf - the buffer to close
 *    clean_mmap - a flag specifying whether or not the mmap'd region should
 *                 be closed.
 * OUTPUTS:
 *    none
 * RETURNS:
 *    none
 * RESTRICTIONS:
 */
void filebuf_close_buffer(filebuf_t *buf, int clean_mmap)
{
#ifndef FILE_MMAP
    if (buf->inbuf)
        FREE(buf->inbuf);
#else
#ifdef FILE_UNIX_MMAP
    if ( clean_mmap )
        if(munmap(buf->fp,buf->len) == -1) {
//            ereport(LOG_CATASTROPHE, XP_GetAdminStr(DBT_munmapFailedS_), system_errmsg());
		;
        }
#else /* FILE_WIN32_MMAP */
    if ( clean_mmap ) {
        UnmapViewOfFile(buf->fp);
        CloseHandle(buf->fdmap);
    }
#endif /* FILE_WIN32_MMAP */
#endif
    FREE(buf);
}

#ifdef FILE_MMAP

#include <sys/stat.h>
/* Shut up compiler about stupid MCOM define */
#ifdef PRIVATE
#undef PRIVATE
#endif

#ifdef FILE_UNIX_MMAP

/* FUNCTION: filebuf_open_nostat
 * DESCRIPTION:
 *    A slightly cheaper version of filebuf_open for if the stat structure
 *    is already available.
 * INPUTS:
 *    fd - the file to open a buffer for
 *    sz - the size of the buffer
 *    finfo - the stat structure for this file, with the length
 *            of the file already known
 * OUTPUTS:
 *    none
 * RETURNS:
 *    NULL on failure
 *    a valid pointer on success
 * RESTRICTIONS:
 */
NSAPI_PUBLIC filebuf_t *filebuf_open_nostat(SYS_FILE fd, int sz, 
                                          struct stat *finfo) 
{
    filebuf_t *buf = (filebuf_t *) MALLOC(sizeof(filebuf_t));

    buf->pos = 0;
    buf->len = finfo->st_size;
    buf->cursize = finfo->st_size;
    buf->fd = fd;

    if(buf->len) {
        buf->fp = (caddr_t)mmap(NULL, buf->len, PROT_READ, FILE_MMAP_FLAGS, PR_FileDesc2NativeHandle(fd), 0);
        if(buf->fp == (caddr_t) -1)
            return NULL;
    }
    buf->inbuf = NULL;
    buf->errmsg = NULL;

    return buf;
}
#else /* FILE_WIN32_MMAP */
NSAPI_PUBLIC filebuf_t *filebuf_open_nostat(SYS_FILE fd, int sz, 
                                          struct stat *finfo) 
{
    filebuf_t *buf = (filebuf_t *) MALLOC(sizeof(filebuf_t));

    buf->pos = 0;
    if (finfo) {
        buf->len = finfo->st_size;
        buf->cursize = finfo->st_size;
    } else {
    	buf->len = buf->cursize = 0;
    }
    buf->fd = fd;

    buf->fdmap = CreateFileMapping((HANDLE)PR_FileDesc2NativeHandle(fd), NULL, PAGE_READONLY, 0, 0, NULL);
    if(!buf->fdmap)
        return NULL;
    if(!(buf->fp = (char *)MapViewOfFile(buf->fdmap, FILE_MAP_READ, 0, 0, 0)))
        return NULL;
    buf->inbuf = NULL;
    buf->errmsg = NULL;

    return buf;
}
#endif /* FILE_*_MMAP */


NSAPI_PUBLIC filebuf_t *filebuf_open(SYS_FILE fd, int sz) 
{
    struct stat finfo;
    PRFileInfo pr_info;

    PR_GetOpenFileInfo(fd, &pr_info);
    finfo.st_mode = 0;
    finfo.st_size = pr_info.size;
    LL_L2I(finfo.st_ctime, pr_info.creationTime);
    LL_L2I(finfo.st_mtime, pr_info.modifyTime);
    
    return filebuf_open_nostat(fd, sz, &finfo);
}

#else   /* !MMAP */

NSAPI_PUBLIC filebuf_t *filebuf_open(SYS_FILE fd, int sz) 
{
    filebuf_t *buf = (filebuf_t *) MALLOC(sizeof(filebuf_t));

    buf->pos = sz;
    buf->cursize = sz; /* force buffer_next on first getc */
    buf->maxsize = sz;
    buf->fd = fd;

    buf->inbuf = NULL;
    buf->errmsg = NULL;

    return buf;
}

#endif


/* ----------------------------- buffer_close ----------------------------- */

#ifdef FILE_UNIX_MMAP
NSAPI_PUBLIC void filebuf_close(filebuf_t *buf) 
{
    if(buf->len) {
        if(munmap(buf->fp,buf->len) == -1) {
            ereport(LOG_CATASTROPHE, XP_GetAdminStr(DBT_munmapFailedS_1), system_errmsg());
            return;
        }
    }
    if(system_fclose(buf->fd) == -1) {
        ereport(LOG_CATASTROPHE, XP_GetAdminStr(DBT_closeFailedS_), system_errmsg());
    }
    FREE(buf);
}

#elif defined(FILE_WIN32_MMAP)
NSAPI_PUBLIC void filebuf_close(filebuf_t *buf) 
{
    UnmapViewOfFile(buf->fp);
    CloseHandle(buf->fdmap);
    system_fclose(buf->fd);

    FREE(buf);
}

#else  /* !MMAP */

NSAPI_PUBLIC void filebuf_close(filebuf_t *buf) {
    system_fclose(buf->fd);
    if(buf->inbuf)
        FREE(buf->inbuf);
    FREE(buf);
}

#endif


/* ----------------------------- buffer_next ------------------------------ */


#ifndef FILE_MMAP

NSAPI_PUBLIC int filebuf_next(filebuf_t *buf, int advance) {
    int n;

    if(!buf->inbuf)
        buf->inbuf = (unsigned char *) MALLOC(buf->maxsize);

    switch(n = system_fread(buf->fd,(char *)(buf->inbuf),buf->maxsize)) {
      case IO_EOF:
        return IO_EOF;
      case IO_ERROR:
        buf->errmsg = system_errmsg(buf->fd);
        return IO_ERROR;
      default:
        buf->pos = advance;
        buf->cursize = n;
        return buf->inbuf[0];
    }
}

#endif



/* ----------------------------- buffer_grab ------------------------------ */


#ifdef FILE_MMAP

NSAPI_PUBLIC int filebuf_grab(filebuf_t *buf, int sz) {
    register int nsz;

    if(buf->len == buf->pos)
        return IO_EOF;

    buf->inbuf = (unsigned char *)&buf->fp[buf->pos];

    nsz = buf->len - buf->pos;
    nsz = (nsz > sz ? sz : nsz);
    
    buf->cursize = nsz;
    buf->pos += nsz;

    return nsz;
}

#else

NSAPI_PUBLIC int filebuf_grab(filebuf_t *buf, int sz) {
    int n;

    if(!buf->inbuf) {
        buf->inbuf = (unsigned char *) MALLOC(sz);
        buf->maxsize = sz;
    }
    else if(sz > buf->maxsize) {
        buf->inbuf = (unsigned char *) REALLOC(buf->inbuf, sz);
        buf->maxsize = sz;
    }

    switch(n = system_fread(buf->fd,(char *)(buf->inbuf),sz)) {
      case IO_EOF:
        return IO_EOF;
      case IO_ERROR:
        buf->errmsg = system_errmsg(buf->sd);
        return IO_ERROR;
      default:
        buf->pos = 0;
        buf->cursize = n;
        return n;
    }
}

#endif



/* ---------------------------- filebuf_buf2sd ---------------------------- */

/*
 * MOVED TO NETBUF.C TO AVOID ENTANGLEMENT IN ADMIN FORMS
 */
